
;--------------------------------------------------------------------------------------------

        ORG     78h

;INT 1Eh vector

BD0078  LABEL   DWORD

;------------------------------------

        ORG     500h

;disk read buffer for root directory

BB0500  LABEL   BYTE

;------------------------------------

        ORG     700h

;disk read buffer for IO.SYS
;this is the address where IO.SYS is eventually executed at 0070:0000

BB0700  LABEL   BYTE

;------------------------------------

        ORG     7C00h

;"bottom" of stack

BW7C00  LABEL   WORD

;branch to executable code

BJ7C00: JMP Short BJ7C3E
        NOP

;------------------------------------

        ORG     7C03h

;OEM identifier

BB7C03  DB 'MSDOS5.0'                    ;8 characters

;------------------------------------

        ORG     7C0Bh

;BIOS Parameter Block (BPB)

;                                         360 Kb   1.2 Mb   720 Kb   1.44 Mb  2.88 Mb

BW7C0B  DW 512   ;bytes per sector         512      512      512      512      512
BB7C0D  DB 2     ;sectors per cluster       2        1        2        1        2
BW7C0E  DW 1     ;reserved sectors          1        1        1        1        1
BB7C10  DB 2     ;number of FAT's           2        2        2        2        2
BW7C11  DW 112   ;root directory entries   112      224      112      224      448
BW7C13  DW 1440  ;total sectors            720      2400     1440     2880     5760
BB7C15  DB 0F9h  ;media descriptor         0FDh     0F9h     0F9h     0F0h     0F0h

;BPB extension 1

BW7C16  DW 3     ;sectors per FAT           2        7        3        9        9
BW7C18  DW 9     ;sectors per head          9       15        9       18       36
BW7C1A  DW 2     ;heads per cylinder        2        2        2        2        2
BW7C1C  DW 0,0   ;hidden sectors            0        0        0        0        0

;BPB extension 2

BD7C20  DD 0                             ;total sectors big (used for hard drives)
BB7C24  DB 0                             ;drive number (80h for hard drives)
BB7C25  DB 0                             ;reserved (DOS:head number NT:dirty)
BB7C26  DB 29h                           ;extended BPB signature
BD7C27  DD 0                             ;volume serial number
BB7C2B  DB 'NO NAME',4 DUP(20h)          ;volume label - 11 characters
BB7C36  DB 'FAT12',3 DUP(20h)            ;file system - 8 characters

;------------------------------------

        ORG     7C3Eh

;the bootstrap routine overwrites part of the code with the original
;BIOS Disk Parameter Table (DPT) and some calculated variables

;the DPT is similar to the BPB with a few changes, including
;some timing values for the Floppy Disk Controller (FDC)

BB7C3E  DB 11 DUP(?)                     ;copy of BIOS DPT
BW7C49  DW ?,?                           ;1st data sector
BW7C4D  DW ?                             ;cylinder number
BB7C4F  DB ?                             ;sector number
BW7C50  DW ?,?                           ;1st root directory sector

;--------------------------------------------------------------------------------------------

        ORG     7C3Eh

;bootstrap code

;set the stack pointer to 0000:7C00

BJ7C3E: CLI                              ;disable maskable interrupts
        XOR     AX,AX                    ;AX = 0
        MOV     SS,AX                    ;SS = 0
        MOV     SP,Offset BW7C00         ;"bottom" of stack

;point to INT 1Eh vector (BIOS DPT pointer)

        PUSH    SS
        POP     ES                       ;ES = 0
        MOV     BX,Offset BD0078         ;INT 1Eh vector offset
        LDS     SI,SS:[BX]               ;DS:SI = pointer to original INT 1Eh DPT

;store the original DPT pointer on the stack, as well as the INT 1Eh vector

        PUSH    DS                       ;DPT pointer
        PUSH    SI
        PUSH    SS                       ;INT 1Eh vector
        PUSH    BX

;copy the original DPT

        MOV     DI,Offset BB7C3E         ;address of BIOS DPT copy
        MOV     CX,11                    ;copy 11 bytes of DPT
        CLD
        REP     MOVSB

;modify the copy

        PUSH    ES
        POP     DS                       ;DS = 0
        MOV Byte Ptr [DI-2],15           ;DPT[9] = FDC timing value
        MOV     CX,BW7C18                ;sectors per head
        MOV     [DI-7],CL                ;DPT[4] = sectors per head

;change the INT 1Eh vector to point to the new DPT

        MOV     [BX+2],AX                ;new INT 1Eh segment = 0
        MOV Word Ptr [BX],Offset BB7C3E  ;new INT 1Eh offset = 7C3Eh

;reset the disk system

        STI                              ;enable maskable interrupts
        INT     13h                      ;reset disk system
        JC      BJ7CED

;here, we calculate the location of the first root directory entry
;so that we may determine if the disk has IO.SYS and MSDOS.SYS on it

        XOR     AX,AX
        CMP     AX,BW7C13                ;total sectors
        JZ      BJ7C84

        MOV     CX,BW7C13                ;total sectors
        MOV     BD7C20,CX                ;total sectors big

BJ7C84: MOV     AL,BB7C10                ;number of FAT's
        MUL Word Ptr BW7C16              ;sectors per FAT
        ADD     AX,BW7C1C                ;hidden sectors
        ADC     DX,BW7C1C+2
        ADD     AX,BW7C0E                ;reserved sectors
        ADC     DX,0                     ;DX:AX = 1st root directory sector
        MOV     BW7C50,AX                ;1st root directory sector
        MOV     BW7C50+2,DX
        MOV     BW7C49,AX                ;initialize 1st data sector value
        MOV     BW7C49+2,DX
        MOV     AX,32                    ;32 bytes per directory entry
        MUL Word Ptr BW7C11              ;root directory entries

;DX:AX = number of bytes in root directory

        MOV     BX,BW7C0B                ;bytes per sector
        ADD     AX,BX
        DEC     AX
        DIV     BX

;AX = number of sectors in root directory

        ADD     BW7C49,AX                ;complete 1st data sector calculation
        ADC Word Ptr BW7C49+2,0

;read the first sector of the root directory into 0000:0500

        MOV     BX,Offset BB0500         ;root directory read buffer
        MOV     DX,BW7C50+2              ;1st root directory sector
        MOV     AX,BW7C50
        CALL    BC7D60                   ;calculate C/H/S values
        JC      BJ7CED

        MOV     AL,1                     ;number of sectors = 1
        CALL    BC7D81                   ;read from disk
        JC      BJ7CED

;compare the first 2 filenames to see if they are IO.SYS and MSDOS.SYS

        MOV     DI,BX
        MOV     CX,11                    ;compare 11 byte filename
        MOV     SI,Offset BB7DE6         ;point to IO.SYS filename
        REPZ    CMPSB
        JNZ     BJ7CED                   ;not a bootable disk

        LEA     DI,[BX+32]               ;point to next directory entry
        MOV     CX,11                    ;compare 11 byte filename
        REPZ    CMPSB
        JZ      BJ7D05                   ;both files exist - load part of IO.SYS

;the files are not present - display error message, wait for a keypress, and reboot

BJ7CED: MOV     SI,Offset BB7D9E         ;non-bootable disk message
        CALL    BC7D52                   ;display a string
        XOR     AX,AX                    ;wait for a keypress
        INT     16h                      ;BIOS keyboard service
        POP     SI                       ;recall INT 1Eh vector
        POP     DS
        POP     [SI]                     ;restore the original DPT pointer
        POP     [SI+2]
        INT     19h                      ;BIOS reboot vector

;error during C/H/S calculation
;toss out the saved stack values and reboot
;a 0 word is left on the stack

BJ7D00: POP     AX
        POP     AX
        POP     AX
        JMP     BJ7CED

;IO.SYS and MSDOS.SYS are present - load the first part of IO.SYS and execute it

BJ7D05: MOV     AX,[BX+26]               ;IO.SYS starting cluster number
        DEC     AX
        DEC     AX
        MOV     BL,BB7C0D                ;sectors per cluster
        XOR     BH,BH
        MUL     BX
        ADD     AX,BW7C49                ;1st data sector
        ADC     DX,BW7C49+2

;this loop reads the first 3 sectors of the bootloader code into 0000:0700 (i.e. 0070:0000)
;after that, the bootloader takes over and is responsible for reading the rest of itself
;this may vary with different versions of DOS

        MOV     BX,Offset BB0700         ;IO.SYS read buffer
        MOV     CX,3                     ;read 3 sectors

BJ7D20: PUSH    AX
        PUSH    DX
        PUSH    CX
        CALL    BC7D60                   ;calculate C/H/S values
        JC      BJ7D00

        MOV     AL,1                     ;number of sectors = 1
        CALL    BC7D81                   ;read from disk
        POP     CX
        POP     DX
        POP     AX
        JC      BJ7CED

        ADD     AX,1
        ADC     DX,0
        ADD     BX,BW7C0B                ;bytes per sector
        LOOP    BJ7D20

;let the bootloader (IO.SYS in this case) take over
;in a real asm file, we would have a segment at 70h and a FAR branch label for 0070:0000

        MOV     CH,BB7C15                ;CH = media descriptor
        MOV     DL,BB7C24                ;DL = drive number
        MOV     BX,BW7C49                ;AX:BX = 1st data sector
        MOV     AX,BW7C49+2
        JMP     0070:0000                ;execute bootloader

;--------------------------------------------------------------------------------------------

BC7D52  PROC    NEAR

;display a zero-terminated string

BJ7D52: LODSB                            ;get the byte from the string
        OR      AL,AL                    ;is it a null terminator ?
        JZ      BJ7D80                   ;yes - exit

        MOV     AH,0Eh                   ;function = write character
        MOV     BX,7                     ;attribute = white on black
        INT     10h                      ;BIOS display service
        JMP     BJ7D52                   ;repeat until null terminator

BC7D52  ENDP

;--------------------------------------------------------------------------------------------

BC7D60  PROC    NEAR

;calculate C/H/S values from an absolute sector value

        CMP     DX,BW7C18                ;sectors per head
        JAE     BJ7D7F

        DIV Word Ptr BW7C18              ;sectors per head
        INC     DL
        MOV     BB7C4F,DL                ;sector number
        XOR     DX,DX
        DIV Word Ptr BW7C1A              ;heads per cylinder
        MOV     BB7C25,DL                ;head number
        MOV     BW7C4D,AX                ;cylinder number
        CLC
        RET

BJ7D7F: STC

BJ7D80:: RET

BC7D60  ENDP

;--------------------------------------------------------------------------------------------

BC7D81  PROC    NEAR

;read from disk

        MOV     AH,2                     ;function = read sector(s)
        MOV     DX,BW7C4D                ;cylinder number
        MOV     CL,6
        SHL     DH,CL
        OR      DH,BB7C4F                ;sector number
        MOV     CX,DX
        XCHG    CH,CL
        MOV     DL,BB7C24                ;drive number
        MOV     DH,BB7C25                ;head number
        INT     13h                      ;BIOS disk I/O service
        RET

BC7D81  ENDP

;--------------------------------------------------------------------------------------------

;string data area

;non-bootable disk message - zero terminated

BB7D9E  DB 13,10,'Non-System disk or disk error',13,10
        DB 'Replace and press any key when ready',13,10,0

;filenames - 11 characters each

BB7DE6  DB 'IO',6 DUP(20h),'SYS'
BB7DF1  DB 'MSDOS',3 DUP(20h),'SYS'

;--------------------------------------------------------------------------------------------

        ORG     7DFEh

;boot sector signature

BW7DFE  DW 0AA55h

;--------------------------------------------------------------------------------------------


